import subprocess

from enum import Enum

from norsu.exceptions import ProcessError


class ExecOutput(Enum):
    """
    Possible outputs of a spawned subprocess.
    """

    Stdout = None
    Pipe = subprocess.PIPE
    Devnull = subprocess.DEVNULL


def execute(args,
            error: bool = True,
            output: ExecOutput = ExecOutput.Pipe,
            **kwargs):
    p = subprocess.Popen(args,
                         stdout=output.value,
                         stderr=subprocess.STDOUT,
                         **kwargs)

    if output == ExecOutput.Pipe:
        out, _ = p.communicate()
        out = out.decode('utf8')
    else:
        p.wait()
        out = None

    if p.returncode != 0:
        if error:
            raise ProcessError('Failed to execute {}'.format(' '.join(args)),
                               stderr=out)  # attach output if possible

    return out



import os
import subprocess
import sys

from distutils.spawn import find_executable
from shutil import rmtree

from norsu.exceptions import LogicError
from norsu.extension import Extension
from norsu.git import find_relevant_refs

from norsu.config import (
    NORSU_DIR,
    WORK_DIR,
    CONFIG,
)

from norsu.instance import (
    Instance,
    InstanceName,
    sort_refs,
    run_temp,
)

from norsu.terminal import (
    Style,
    give_terminal_to,
)

from norsu.utils import (
    partition,
    str_args_to_dict,
)


def known_targets(directory=NORSU_DIR):
    return {e for e in os.listdir(directory) if not e.startswith('.')}


def preprocess_targets(raw_targets, directory=NORSU_DIR):
    entries_pos, entries_neg = partition(lambda x: x.startswith('^'),
                                         raw_targets)

    entries_neg = set((e[1:] for e in entries_neg))  # remove '^'
    entries_pos = set(entries_pos)

    if not entries_pos or entries_neg:
        entries_pos = known_targets(directory=directory)

    entries = []
    for e in sorted(entries_pos - entries_neg):
        name, _, query = e.partition(':')
        entries.append(InstanceName(name=name, query=query))

    return entries


def split_make_args(args):
    entries, options = partition(lambda x: x.startswith('-'), args)

    entries = list(entries)
    options = list(options)

    return entries, options


def cmd_install(args, _):
    for target in preprocess_targets(args.target):
        print('Selected instance:', Style.bold(target))

        Instance(target).install(configure=args.configure,
                                 extensions=args.extensions,
                                 update=not args.no_update)

        print()  # splitter


def cmd_instance(args, _):
    cmd = args.command

    # Protect the user from an accidental removal of all instances
    if cmd == 'remove' and not (args.target or args.force):
        raise LogicError('Pass --force to remove all instances')

    for target in preprocess_targets(args.target):
        print('Selected instance:', Style.bold(target))

        instance = Instance(target)

        cmds = {
            'pull': lambda: instance.pull(),
            'remove': lambda: instance.remove(),
            'status': lambda: instance.status(),
        }

        # execute command
        cmds[cmd]()

        print()  # splitter


def cmd_run(main_args, cli_args):
    instance = Instance(main_args.target)
    dbname = main_args.dbname
    port = main_args.port

    dump_file = main_args.dump
    restore_file = main_args.restore

    work_dir = os.getcwd()
    pg_config = instance.get_bin_path('pg_config')

    config_files = []

    # Grab extra extension options
    if main_args.pgxs:
        extension = Extension(work_dir=work_dir, pg_config=pg_config)

        mk_var = 'EXTRA_REGRESS_OPTS'
        regress_opts = str_args_to_dict(extension.makefile_var(mk_var))
        config_files.append(regress_opts.get('--temp-config'))

    # Grab extra PG config files
    if main_args.config:
        config_files.extend(main_args.config)

    with run_temp(instance, config_files=config_files, port=port) as node:
        if restore_file:
            node.restore(restore_file, dbname=dbname)
            print('Restored from', Style.bold(restore_file))

        cli = (main_args.psql and 'psql') or main_args.interactive
        if cli:
            # Either it's a binary present in this instance
            # or it should be in PATH
            exe = instance.get_bin_path(cli)
            if not os.path.exists(exe):
                exe = find_executable(cli)
            cmd = [
                exe,
                f'postgres://localhost:{node.port}/{dbname}',
                *cli_args,
            ]
            print(' '.join(cmd))

        else:
            print('Press Ctrl+C to exit', file=sys.stderr)
            cmd = ['sleep', 'infinity']

        # XXX: to avoid undesirable PG shutdown (signal),
        # run cmd within its own process group
        p = subprocess.Popen(cmd, preexec_fn=os.setpgrp)

        # give PTS control to cmd and wait for it to finish
        give_terminal_to(p.pid)
        p.wait()

        if dump_file:
            filename = node.dump(filename=dump_file, dbname=dbname)
            print('Dump has been saved to', Style.bold(filename))

        sys.exit(p.returncode)


def cmd_search(args, _):
    for target in preprocess_targets(args.target):
        print('Search query:', Style.bold(target.query),
              f'({target.type.name})')

        patterns = target.to_patterns()
        refs = find_relevant_refs(CONFIG['repos']['urls'], patterns)

        for ref in sort_refs(refs, target):
            print('\t', ref.name)

        print()  # splitter


def cmd_purge(args, _):
    for target in preprocess_targets(args.target, WORK_DIR):
        instance = Instance(target)
        if not os.path.exists(instance.main_dir):
            rmtree(path=instance.work_dir, ignore_errors=True)


def cmd_pgxs(main_args, make_args):
    make_targets, make_opts = split_make_args(make_args)
    work_dir = os.getcwd()

    for pg in preprocess_targets(main_args.target):
        instance = Instance(pg)
        pg_config = instance.get_bin_path('pg_config')
        extension = Extension(work_dir=work_dir, pg_config=pg_config)

        if os.path.exists(pg_config):
            print('Executing against instance', Style.bold(pg), '\n')
        else:
            print(Style.yellow(f'Cannot find instance {pg}\n'))
            continue

        # should we start PostgreSQL?
        if main_args.run_pg:
            port = main_args.run_pg_port

            mk_var = 'EXTRA_REGRESS_OPTS'
            regress_opts = str_args_to_dict(extension.makefile_var(mk_var))
            config_files = [regress_opts.get('--temp-config')]

            # run commands under a running PostgreSQL instance
            with run_temp(instance, config_files=config_files,
                          port=port) as node:
                # make pg_regress aware of non-default port
                make_opts.append(f'{mk_var}+=--port={node.port}')
                extension.make(*make_targets, options=make_opts)
        else:
            extension.make(*make_targets, options=make_opts)

        print()  # splitter


def cmd_path(args, _):
    for target in preprocess_targets(args.target):
        print(Instance(target).main_dir)



import os
import signal

from norsu.config import CONFIG


class Style:
    @staticmethod
    def style(color, text):
        if os.isatty(1) and os.isatty(2) and CONFIG['misc']['colors']:
            return f'\033[{color}m{text}\033[0m'
        return text

    @staticmethod
    def bold(text):
        return Style.style(1, text)

    @staticmethod
    def red(text):
        return Style.style(31, text)

    @staticmethod
    def green(text):
        return Style.style(32, text)

    @staticmethod
    def blue(text):
        return Style.style(94, text)

    @staticmethod
    def yellow(text):
        return Style.style(33, text)


def give_terminal_to(pgid):
    signals = {
        signal.SIGTTOU,
        signal.SIGTTIN,
        signal.SIGTSTP,
        signal.SIGCHLD,
    }

    old_mask = signal.pthread_sigmask(signal.SIG_BLOCK, signals)
    try:
        os.tcsetpgrp(2, pgid)
        return True
    except ProcessLookupError:
        return False
    except OSError:
        return False
    finally:
        signal.pthread_sigmask(signal.SIG_SETMASK, old_mask)



import argparse
import shlex

from typing import List


class ShlexSplitAction(argparse.Action):
    def __init__(self, option_strings, dest, **kwargs):
        super(ShlexSplitAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, shlex.split(values))


def split_args_extra(args: List[str]):
    """
    Separate main args from auxiliary ones.
    """

    try:
        i = args.index('--')
        return args[:i], args[i + 1:]
    except ValueError:
        return args, []



class ProcessError(Exception):
    """
    This exception represents a failed invocation of an external tool.
    """
    def __init__(self, message='', stderr=None):
        super().__init__(message)
        self.stderr = stderr


class LogicError(Exception):
    """
    Main application error.
    """



import os
import shlex
import importlib

from norsu.config import CONFIG, TOOL_MAKE
from norsu.exceptions import LogicError, ProcessError
from norsu.execute import ExecOutput, execute
from norsu.terminal import Style


class Extension:
    def __init__(self, work_dir, pg_config=None):
        self.work_dir = work_dir
        self.pg_config = pg_config

    def __pgxs_args(self):
        if self.pg_config:
            return ['USE_PGXS=1', f'PG_CONFIG={self.pg_config}']
        return []

    def make(self, *targets, options=None):
        if not targets:
            targets = CONFIG['pgxs']['default_targets']

        if not options:
            options = CONFIG['pgxs']['default_options']

        # copy options
        opts = options[:]

        # append compiler options, if needed (e.g. for scan_build)
        for env in ['CC', 'CXX']:
            if env in os.environ:
                opts.append(f'{env}={os.environ.get(env)}')

        for target in targets:
            # print simplified command
            quoted_opts = ' '.join([shlex.quote(x) for x in opts])
            s = f'$ make {quoted_opts} {target}'
            print(Style.green(s))

            args = [
                TOOL_MAKE,
                *self.__pgxs_args(),
                *opts,
                target,
            ]

            # execute make (writes to stdout)
            execute(args,
                    cwd=self.work_dir,
                    env=os.environ,
                    output=ExecOutput.Stdout)
            print()

    def makefile_var(self, name):
        makefile = os.path.join(self.work_dir, 'Makefile')
        print_mk = importlib.resources.path('norsu', 'data/print.mk')

        args = [
            TOOL_MAKE,
            *self.__pgxs_args(),
            '-f',
            makefile,
            '-f',
            print_mk,
            f'print-{name}',
        ]

        try:
            # return var's value
            return execute(args).partition('=')[2]
        except ProcessError as e:
            raise LogicError(
                f'Failed to get variable {name} from Makefile') from e



import os
import shlex
import sys

from itertools import tee, filterfalse
from typing import Dict, Optional


def partition(pred, iterable):
    t1, t2 = tee(iterable)
    return filterfalse(pred, t1), filter(pred, t2)


def str_args_to_dict(a: str) -> Dict[str, str]:
    result = {}
    for arg in shlex.split(a):
        k, _, v = arg.partition('=')
        result[k] = v
    return result


def path_exists(path: str) -> Optional[str]:
    if os.path.exists(path):
        return path


def eprint(*args, **kwargs) -> None:
    print(*args, file=sys.stderr, **kwargs)


def limit_lines(string: str, n: int) -> str:
    return '\n'.join(string.splitlines()[-n:])



import argparse
import sys

import norsu.commands as commands

from norsu import __version__
from norsu.exceptions import LogicError, ProcessError
from norsu.terminal import Style

from norsu.args import (
    ShlexSplitAction,
    split_args_extra,
)

from norsu.utils import (
    eprint,
    limit_lines,
)


def main():
    # split args using '--'
    args, extra = split_args_extra(sys.argv)

    app = args[0]
    examples = f"""
examples:
    {app} install  9.6.5  10  master
    {app} pgxs     9.6   9.5  --  install -j4
    {app} pull     REL_10_STABLE
    {app} remove   9.5
    {app} status
    """

    # Hint for all TARGET args
    known_targets = commands.known_targets()

    parser = argparse.ArgumentParser(
        description=f'PostgreSQL builds manager v{__version__}',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=examples)

    parser.set_defaults(func=None)

    subparsers = parser.add_subparsers(title='commands', dest='command')

    # norsu install
    p_install = subparsers.add_parser(
        'install', description='build & install a list of versions')
    p_install.add_argument('target', nargs='*')
    p_install.add_argument('--configure',
                           action=ShlexSplitAction,
                           help='options for ./configure')
    p_install.add_argument('--extensions',
                           nargs='*',
                           help='also install listed exceptions')
    p_install.add_argument('-E',
                           '--no-update',
                           action='store_true',
                           help='do not pull and install updates')
    p_install.set_defaults(func=commands.cmd_install)

    # norsu remove
    p_remove = subparsers.add_parser('remove',
                                     description='remove specified builds')
    p_remove.add_argument('target', nargs='*')
    p_remove.add_argument('--force',
                          action='store_true',
                          help='force remove, even if no target')
    p_remove.set_defaults(func=commands.cmd_instance)

    # norsu status
    p_status = subparsers.add_parser(
        'status', description='show some info for each build installed')
    p_status.add_argument('target', nargs='*')
    p_status.set_defaults(func=commands.cmd_instance)

    # norsu pull
    p_pull = subparsers.add_parser(
        'pull', description='pull latest changes from git repos')
    p_pull.add_argument('target', nargs='*')
    p_pull.set_defaults(func=commands.cmd_instance)

    # norsu search
    p_search = subparsers.add_parser(
        'search', description='find matching branches in git repos')
    p_search.add_argument('target', nargs='*')
    p_search.set_defaults(func=commands.cmd_search)

    # norsu purge
    p_purge = subparsers.add_parser(
        'purge', description='remove orphaned cloned git repos')
    p_purge.add_argument('target', nargs='*')
    p_purge.set_defaults(func=commands.cmd_purge)

    # norsu pgxs
    p_pgxs = subparsers.add_parser(
        'pgxs', description='run "make USE_PGXS=1 ..." in current dir')
    p_pgxs.add_argument('target', nargs='*')
    p_pgxs.add_argument('-R',
                        '--run-pg',
                        action='store_true',
                        help='run temp instance')
    p_pgxs.add_argument('--run-pg-port',
                        type=int,
                        help='port to be used for temp instance')
    p_pgxs.set_defaults(func=commands.cmd_pgxs)

    # norsu run
    p_run = subparsers.add_parser(
        'run', description='run a temp instance of PostgreSQL')
    p_run.add_argument('target', choices=known_targets)
    p_run.add_argument('--config',
                       nargs='*',
                       help='additional config files for PostgreSQL')
    p_run.add_argument('--psql',
                       action='store_true',
                       help='[DEPRECATED] run PSQL after PG has started')
    p_run.add_argument('-i',
                       '--interactive',
                       action='store',
                       const='psql',
                       metavar='CLI',
                       nargs='?',
                       help='start an interactive CLI connected to a DB')
    p_run.add_argument('--pgxs',
                       action='store_true',
                       help='grab PGXS config as well')
    p_run.add_argument('--dbname',
                       default='postgres',
                       help='database name for PSQL')
    p_run.add_argument('--port',
                       type=int,
                       help='port to be used for this instance')
    p_run.add_argument('--dump',
                       metavar='FILENAME',
                       type=str,
                       help='dump a DB to a file before shutdown')
    p_run.add_argument('--restore',
                       metavar='FILENAME',
                       type=str,
                       help='restore a DB from a file')
    p_run.set_defaults(func=commands.cmd_run)

    # norsu path
    p_path = subparsers.add_parser(
        'path', description='show paths to a specific build')
    p_path.add_argument('target', nargs='*')
    p_path.set_defaults(func=commands.cmd_path)

    try:
        parsed_args = parser.parse_args(args[1:])
        command = parsed_args.func

        if command:
            # execute a suitable command
            command(parsed_args, extra)
        else:
            parser.print_help()

    except KeyboardInterrupt:
        pass

    except LogicError as e:
        eprint(Style.red(str(e)))
        sys.exit(1)

    except ProcessError as e:
        eprint(Style.red(str(e)))
        if e.stderr:
            eprint('LOG:\n\n<... skipped lines ...>')
            eprint(limit_lines(e.stderr, 8))
        sys.exit(1)

    # XXX: We deliberately don't catch Exception,
    # since it might point to application's bugs
    except Exception:
        raise



import os
import re
import shlex
import sys
import multiprocessing

from contextlib import contextmanager, redirect_stdout
from enum import Enum
from shutil import rmtree
from testgres import get_new_node, configure_testgres

from norsu.config import NORSU_DIR, WORK_DIR, CONFIG, TOOL_MAKE
from norsu.exceptions import LogicError, ProcessError
from norsu.execute import ExecOutput, execute
from norsu.extension import Extension
from norsu.terminal import Style

from norsu.git import (
    GitRepo,
    SortRefBySimilarity,
    SortRefByVersion,
    find_relevant_refs,
)

from norsu.utils import (
    eprint,
    path_exists,
)


def step(*args):
    print(Style.green('\t=>'), *args)


def line(name, value=None):
    print('\t', name, f'\t{value}')


def read_commit_file(path):
    if os.path.exists(path):
        with open(path, 'r') as f:
            return f.read().strip()


def write_commit_file(path, value):
    with open(path, 'w') as f:
        f.write(value or '')


def sort_refs(refs, name):
    # key function for sort
    def to_key(x):
        if name.type == InstanceNameType.Version:
            return SortRefByVersion(x)
        else:
            # pre-calculated for better performance
            name_ngram = SortRefBySimilarity.ngram(name.query)
            return SortRefBySimilarity(x, name_ngram)

    return sorted(refs, reverse=True, key=to_key)


class InstanceNameType(Enum):
    Version = 1
    Branch = 2


class InstanceName:
    rx_is_ver = re.compile(r'\d+([._]\d+)*')
    rx_sep = re.compile(r'(\.|_)')

    @staticmethod
    def _check_str(s):
        pred1 = len(s.strip()) > 0
        pred2 = any(c.isalnum() for c in s)

        if not (pred1 and pred2):
            raise LogicError('Bad identifier: {s}')

        return s

    def __init__(self, name, query=None):
        self.value = self._check_str(name)
        self.query = self._check_str(query or name)

        if self.rx_is_ver.match(self.query):
            self.type = InstanceNameType.Version
        else:
            self.type = InstanceNameType.Branch

    def to_patterns(self):
        pattern = self.query
        result = [pattern]

        if self.type == InstanceNameType.Version:
            # replace version separators with a pattern
            pattern = self.rx_sep.sub(lambda m: '[._]', pattern)

            result.extend([
                f'REL_{pattern}*',
                f'REL{pattern}*',
            ])
        else:
            result.append(f'*{pattern}*')

        return result

    def __str__(self):
        return self.value


class Instance:
    def __init__(self, name):
        if isinstance(name, InstanceName):
            self.name = name
        else:
            self.name = InstanceName(name)

        self.main_dir = os.path.join(NORSU_DIR, str(name))
        self.work_dir = os.path.join(WORK_DIR, str(name))
        self.git = GitRepo(work_dir=self.work_dir)

        # various utility files
        self.ignore_file = os.path.join(self.main_dir, '.norsu_ignore')

        # store commit hashes (build + install)
        self.installed_commit_file = os.path.join(self.main_dir,
                                                  '.norsu_build')
        self.built_commit_file = os.path.join(self.work_dir, '.norsu_build')

    @property
    def ignore(self):
        return os.path.exists(self.ignore_file)

    @property
    def actual_commit_hash(self):
        if os.path.exists(self.work_dir):
            return self.git.hash

    @property
    def installed_commit_hash(self):
        return read_commit_file(self.installed_commit_file)

    @installed_commit_hash.setter
    def installed_commit_hash(self, value):
        write_commit_file(self.installed_commit_file, value)

    @property
    def built_commit_hash(self):
        return read_commit_file(self.built_commit_file)

    @built_commit_hash.setter
    def built_commit_hash(self, value):
        write_commit_file(self.built_commit_file, value)

    @property
    def requires_reinstall(self):
        # NOTE: remember that re-build != re-install!
        # We might already have a fresh build in work dir
        ic = self.installed_commit_hash
        ac = self.actual_commit_hash

        return ic != ac or not ic or not ac

    @property
    def requires_rebuild(self):
        bc = self.built_commit_hash
        ac = self.actual_commit_hash

        return bc != ac or not bc or not ac

    def get_bin_path(self, name):
        return os.path.join(self.main_dir, 'bin', name)

    def pg_config(self, params=None):
        pg_config = self.get_bin_path('pg_config')
        if os.path.exists(pg_config):
            return execute([pg_config] + params)

    def status(self):
        postgres = os.path.join(self.main_dir, 'bin', 'postgres')

        if os.path.exists(postgres):
            if self.requires_reinstall:
                status = Style.yellow('Installed (out of date)')
            else:
                status = Style.green('Installed')
        else:
            status = Style.red('Not installed')

        line('Status:', status)
        line('Main dir:', path_exists(self.main_dir))
        line('Work dir:', path_exists(self.work_dir))

        if os.path.exists(self.work_dir):
            branch = self.git.branch or self.git.tag
            if branch:
                line('Branch:', branch)

        pg_config_out = self.pg_config(['--version'])
        if pg_config_out:
            line('Version:', pg_config_out.strip())

        commit = self.installed_commit_hash
        if commit:
            line('Commit:', commit)

        pg_config_manual = os.path.join(self.main_dir, 'include',
                                        'pg_config_manual.h')
        if os.path.exists(pg_config_manual):
            with open(pg_config_manual, 'r') as f:
                for ln in f:
                    if ln.startswith('#define MEMORY_CONTEXT_CHECKING'):
                        break  # too late
                    if ln.startswith('#define USE_VALGRIND'):
                        line('Valgrind:', 'Enabled')
                        break  # OK

        configure = self._configure_options()
        line('CONFIGURE:', configure)

    def pull(self):
        if os.path.exists(self.main_dir) and not os.path.exists(self.work_dir):
            step(Style.yellow('This is a standalone build, skipping'))
        else:
            self._maybe_git_clone_or_pull(update=True)

    def install(self, configure=None, extensions=None, update=True):
        if self.ignore:
            step(Style.yellow('Ignored due to .norsu_ignore'))

        elif os.path.exists(
                self.main_dir) and not os.path.exists(self.work_dir):
            step(Style.yellow('This is a standalone build, skipping'))

        else:
            try:
                self._maybe_git_clone_or_pull(update)
                self._maybe_make_distclean(configure)
                self._maybe_configure_project(configure)
                self._maybe_make_install(configure)
                self._maybe_make_extensions(extensions)
            except ProcessError as e:
                step(Style.red(str(e)))

                # We'd like to print log and exit with error code
                raise ProcessError(stderr=e.stderr)

    def remove(self):
        for path, name in [(self.main_dir, 'main'), (self.work_dir, 'work')]:
            if os.path.exists(path):
                rmtree(path=path, ignore_errors=True)
                step(f'Removed directory {name}')

    def _configure_options(self):
        pg_config_out = self.pg_config(['--configure'])
        if pg_config_out:
            options = shlex.split(pg_config_out)
            return [x for x in options if not x.startswith('--prefix')]

        return CONFIG['build']['configure_options']

    def _configure_options_are_new(self, opts):
        # operation's required if new non-trivial configure flags
        return opts is not None and opts != self._configure_options()

    def _maybe_git_clone_or_pull(self, update):
        git_repo = os.path.join(self.work_dir, '.git')

        if not os.path.exists(git_repo):
            step('No work dir, choosing repo & branch')

            patterns = self.name.to_patterns()
            refs = find_relevant_refs(CONFIG['repos']['urls'], patterns)

            if not refs:
                raise LogicError(f'No branch found for {self.name}')

            # select the most relevant branch
            ref = sort_refs(refs, self.name)[0]
            step('Selected repo', Style.bold(ref.repo))
            step('Selected branch', Style.bold(ref.name))

            # finally, clone repo
            self.git.clone(url=ref.repo, branch=ref.name)
            step('Cloned git repo to work dir')

        # Are we allowed to update this instance?
        elif update:
            branch = self.git.branch

            # pull latest changes
            if branch:
                self.git.pull()

            # should we reinstall PG?
            if self.requires_reinstall:
                fresh_commits = ''

                # show distance between installed and fresh commits
                installed_commit = self.installed_commit_hash
                if installed_commit:
                    commits = self.git.distance(installed_commit, branch)
                    fresh_commits = f'({commits} commits)'

                step('Current branch:', Style.bold(branch))
                step('Installed build is out of date', fresh_commits)

        # add .norsu* to git excludes
        self.git.add_excludes('.norsu*')

    def _maybe_configure_project(self, configure):
        makefile = os.path.join(self.work_dir, 'GNUmakefile')
        if not os.path.exists(makefile):
            args = ['./configure', f'--prefix={self.main_dir}']

            # NOTE: [] is a valid choice
            if configure is None:
                configure = self._configure_options()

            if configure:
                args.extend(configure)

            execute(args, cwd=self.work_dir)
            step('Configured sources with', configure)

    def _maybe_make_distclean(self, configure):
        makefile = os.path.join(self.work_dir, 'GNUmakefile')
        new_conf_opts = self._configure_options_are_new(configure)

        if os.path.exists(makefile) and \
           (new_conf_opts or self.requires_rebuild):

            # reset built commit hash
            self.built_commit_hash = None

            args = [TOOL_MAKE, 'distclean']
            execute(args,
                    cwd=self.work_dir,
                    error=False,
                    output=ExecOutput.Devnull)

            step('Prepared work dir for a new build')

    def _maybe_make_install(self, configure):
        new_conf_opts = self._configure_options_are_new(configure)

        if new_conf_opts or self.requires_reinstall:
            # update built commit hash
            self.built_commit_hash = self.actual_commit_hash

            jobs = int(CONFIG['build']['jobs'])
            if jobs == 0:
                 jobs = multiprocessing.cpu_count()

            for arg in [f'-j{jobs}', 'install']:
                args = [TOOL_MAKE, arg]
                execute(args, cwd=self.work_dir)

            # update installed commit hash
            self.installed_commit_hash = self.actual_commit_hash

            step('Built and installed')

    def _maybe_make_extensions(self, extensions=None):
        if extensions is None:
            return

        # provide defaults
        if not extensions:
            path = os.path.join(self.work_dir, 'contrib')
            extensions = sorted((e for e in os.listdir(path)
                                 if os.path.isdir(os.path.join(path, e))))

        failed = False
        for extension in extensions:
            # is it a contrib?
            path = os.path.join(self.work_dir, 'contrib', extension)
            try:
                Extension(path).make('install')
                step('Installed contrib', Style.bold(extension))
            except ProcessError:
                step(Style.red(f'Failed to install {extension}'))
                failed = True

        if failed:
            raise LogicError('Failed to install some extensions')


@contextmanager
def run_temp(instance, config_files=None, **kwargs):
    pg_config = instance.get_bin_path('pg_config')
    temp_conf = ''

    if not os.path.exists(pg_config):
        raise LogicError(f'Failed to find pg_config at {pg_config}')

    # HACK: help testgres find our instance
    os.environ['PG_CONFIG'] = pg_config

    # disable instance caching
    configure_testgres(cache_initdb=False)

    if config_files:
        configs = []

        for path in (f for f in config_files if f is not None):
            eprint('Custom config file:', os.path.basename(path))
            with open(path) as f:
                configs.append(f.read())

        temp_conf = '\n'.join(configs)

    with get_new_node(**kwargs) as node:
        with redirect_stdout(sys.stderr):
            print('Starting temporary PostgreSQL instance...')
            print()
            print('dir:', node.base_dir)
            print('port:', node.port)
            print()

        # prepare and start a new node
        node.cleanup_on_bad_exit = True
        node.init().append_conf(line=temp_conf).start()

        yield node



#!/usr/bin/env python

import norsu
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

install_requires = ['toml', 'testgres']

# Get contents of README file
with open('README.md', 'r') as f:
    readme = f.read()

setup(name='norsu',
      version=norsu.__version__,
      packages=['norsu'],
      package_data={'norsu': ['data/*']},
      license='MIT',
      author='Dmitry Ivanov',
      url='https://github.com/funbringer/norsu',
      long_description=readme,
      long_description_content_type='text/markdown',
      description='PostgreSQL builds manager',
      keywords=['PostgreSQL', 'postgres', 'install', 'test'],
      install_requires=install_requires,
      entry_points={'console_scripts': ['norsu = norsu.main:main']})



import os

from functools import total_ordering

from norsu.config import CONFIG
from norsu.execute import ExecOutput, execute


@total_ordering
class SortRefByVersion:
    def __init__(self, ref):
        # use lowercase for substr search
        name = ref.name.lower()

        # extract version numbers
        ver = ''.join((c for c in name if c.isdigit() or c == '_'))
        ver = (n for n in ver.split('_') if n)
        ver = list(map(lambda n: int(n), ver))

        types = [
            ('stable', float('+inf')),
            ('rc', -1),
            ('beta', -2),
            ('alpha', -3),
        ]

        for t, num in types:
            if t in name:
                # example:
                #  REL_10_RC1 => (10, -1, 1)
                #  REL_9_6_STABLE => (9, 6, 'inf')
                _, _, s = name.rpartition(t)
                if s.isdigit():
                    ver.pop()  # see ver extraction
                    ver.append(num)
                    ver.append(int(s))
                else:
                    ver.append(num)

        self.ref = ref
        self.ver = ver

    def __eq__(self, other):
        return self.ver == other.ver

    def __lt__(self, other):
        return self.ver < other.ver

    def __str__(self):
        return self.ref


@total_ordering
class SortRefBySimilarity:
    @staticmethod
    def ngram(text, n=3):
        ngrams = (text[i:i + n] for i in range(0, len(text) - n + 1))
        return set(ngrams)

    @staticmethod
    def similarity(ng1, ng2):
        return len(ng1 & ng2) / float(len(ng1 | ng2))

    def __init__(self, ref, name_ngram):
        ng1 = self.ngram(ref.name)
        ng2 = name_ngram

        self.similarity = self.similarity(ng1, ng2)
        self.ref = ref

    def __eq__(self, other):
        return self.similarity == other.similarity

    def __lt__(self, other):
        return self.similarity < other.similarity


class GitRef:
    def __init__(self, repo, name):
        self.repo = repo
        self.name = name


class GitRepo:
    def __init__(self, work_dir, url=None):
        self.work_dir = work_dir
        self.url = url

    @property
    def branch(self):
        args = ['git', 'symbolic-ref', '--short', 'HEAD']
        out = execute(args, cwd=self.work_dir, error=False)
        if out:
            return out.strip()

    @property
    def tag(self):
        args = ['git', 'tag', '--points-at', 'HEAD']
        out = execute(args, cwd=self.work_dir, error=False)
        if out:
            return out.strip()

    @property
    def hash(self):
        args = ['git', 'rev-parse', 'HEAD']
        return execute(args, cwd=self.work_dir).strip()

    def clone(self, url=None, branch='master', depth=1):
        args = [
            'git',
            'clone',
            '--branch',
            branch,
            '--depth',
            str(depth),
            url or self.url,
            self.work_dir,
        ]
        execute(args, output=ExecOutput.Devnull)

    def pull(self, remote='origin', branch=None):
        args = ['git', 'pull', remote, branch or self.branch]
        execute(args, cwd=self.work_dir, output=ExecOutput.Devnull)

    def distance(self, commit1, commit2):
        args = [
            'git',
            'rev-list',
            f'{commit1}..{commit2}',
            '--count',
        ]

        try:
            # mute possible cast errors
            return int(execute(args, cwd=self.work_dir).strip())
        except ValueError:
            pass

    def add_excludes(self, pattern):
        excludes = os.path.join(self.work_dir, '.git', 'info', 'exclude')
        with open(excludes, 'r+') as f:
            lines = f.readlines()
            if not any(pattern in s for s in lines):
                f.seek(0, os.SEEK_END)
                f.write(pattern)


def find_relevant_refs(repos, patterns):
    refs = []

    for repo in repos:
        args = ['git', 'ls-remote', '--heads', '--tags', repo]
        args += patterns  # search patterns

        out = execute(args)

        # list of matching branches and tags
        refs += [
            GitRef(repo, os.path.basename(r.split()[-1]))
            for r in out.splitlines()
        ]

        # should we stop after 1st match?
        if refs and CONFIG['repos']['first_match']:
            break

    return refs
